"""
ZarrWriter

This writer will write a single zarr file.
That file can then be opened with zarr.open
"""

import json

import numpy as np
import zarr
from pymmcore_plus import CMMCorePlus
from useq import MDASequence

from pymmcore_mda_writers import ZarrWriter

core = CMMCorePlus.instance()
core.loadSystemConfiguration()

mda = MDASequence(
    time_plan={"interval": 0.1, "loops": 2},
    stage_positions=[(1, 1, 1)],
    z_plan={"range": 3, "step": 1},
    channels=[{"config": "DAPI", "exposure": 1}],
)

# This will automatically get the same core instance we used above
# if you are not using the core singleton make sure to pass core to
# the writer init.

# the zarr writer requires that give it the dtype and shape of images on init
# For the demo camera these will be (512, 512) and uint16 respectively
writer = ZarrWriter("data/zarr_writer_example/run", (512, 512), np.uint16)


# run the MDA twice
# the writer will not overwrite an existing store
# instead it will make a new one automatically

th = core.run_mda(mda)
th.join()

th = core.run_mda(mda)
th.join()

# If you run  may need to increment
run_1 = zarr.open("data/zarr_writer_example/run_1.zarr")
run_2 = zarr.open("data/zarr_writer_example/run_2.zarr")


print(run_2)

# metadata as well as the original MDASequence are stored
# in the zarr `.attrs`

print(run_2.attrs.asdict())

# you can fully reconstruct the original sequence like so
saved_sequence_as_useq = MDASequence(**json.loads(run_2.attrs["useq-sequence"]))
print(saved_sequence_as_useq)



try:
    from ._version import version as __version__
except ImportError:
    __version__ = "unknown"  # pragma: no cover
__author__ = "Ian Hunt-Isaak"
__email__ = "ianhuntisaak@gmail.com"

from ._writers import BaseWriter, SimpleMultiFileTiffWriter, ZarrWriter

__all__ = [
    "__version__",
    "BaseWriter",
    "SimpleMultiFileTiffWriter",
    "ZarrWriter",
]



__all__ = [
    "BaseWriter",
    "SimpleMultiFileTiffWriter",
    "ZarrWriter",
]
from pathlib import Path
from typing import Optional, Sequence, Tuple, Union

import numpy as np
import numpy.typing as npt
from pymmcore_plus import CMMCorePlus
from pymmcore_plus.mda import PMDAEngine
from useq import MDAEvent, MDASequence

try:
    import tifffile
except ModuleNotFoundError:
    tifffile = None
try:
    import zarr
except ModuleNotFoundError:
    zarr = None


class BaseWriter:
    def __init__(self, core: CMMCorePlus = None) -> None:
        self._core = core or CMMCorePlus.instance()
        self._on_mda_engine_registered(self._core.mda, None)
        self._core.events.mdaEngineRegistered.connect(self._on_mda_engine_registered)
        # TODO add paused and finished events

    def _onMDAStarted(self, sequence: MDASequence):
        ...

    def _onMDAFrame(self, img: np.ndarray, event: MDAEvent):
        ...  # pragma: no cover

    def _on_mda_engine_registered(
        self, newEngine: PMDAEngine, oldEngine: Optional[PMDAEngine] = None
    ):
        if oldEngine:
            self._disconnect(oldEngine)
        newEngine.events.sequenceStarted.connect(self._onMDAStarted)
        newEngine.events.frameReady.connect(self._onMDAFrame)

    def _disconnect(self, engine: PMDAEngine):
        engine.events.sequenceStarted.disconnect(self._onMDAStarted)
        engine.events.frameReady.disconnect(self._onMDAFrame)

    def disconnect(self):
        "Disconnect this writer from processing any more events"
        self._disconnect(self._core.mda)

    @staticmethod
    def get_unique_folder(
        folder_base_name: Union[str, Path],
        suffix: Union[str, Path] = None,
        create: bool = False,
    ) -> Path:
        """
        Get a unique foldername of the form '{folder_base_name}_{i}

        Parameters
        ----------
        folder_base_name : str or Path
            The folder name in which to put data
        suffix : str or Path
            If given, to be used as the path suffix. e.g. `.zarr`
        create : bool, default False
            Whether to create the folder.
        '"""
        folder = Path(folder_base_name).resolve()
        stem = str(folder.stem)

        def new_path(i):
            path = folder.parent / (stem + f"_{i}")
            if suffix:
                return path.with_suffix(suffix)
            return path

        i = 1
        path = new_path(i)
        while path.exists():
            i += 1
            path = new_path(i)
        if create:
            path.mkdir(parents=True)
        return path

    @staticmethod
    def sequence_axis_order(seq: MDASequence) -> Tuple[str]:
        """Get the axis order using only axes that are present in events."""
        # hacky way to drop unncessary parts of the axis order
        # e.g. drop the `p` in `tpcz` if there is only one position
        # TODO: add a better implementation upstream in useq
        event = next(seq.iter_events())
        event_axes = list(event.index.keys())
        return tuple(a for a in seq.axis_order if a in event_axes)

    @staticmethod
    def event_to_index(axis_order: Sequence[str], event: MDAEvent) -> Tuple[int, ...]:
        return tuple(event.index[a] for a in axis_order)


class SimpleMultiFileTiffWriter(BaseWriter):
    def __init__(
        self, data_folder_name: Union[str, Path], core: CMMCorePlus = None
    ) -> None:
        if tifffile is None:
            raise ValueError(
                "This writer requires tifffile to be installed. "
                "Try: `pip install tifffile`"
            )
        super().__init__(core)
        self._data_folder_name = data_folder_name

    def _onMDAStarted(self, sequence: MDASequence) -> None:
        self._path = self.get_unique_folder(self._data_folder_name, create=True)
        self._axis_order = self.sequence_axis_order(sequence)
        with open(self._path / "useq-sequence.json", "w") as f:
            f.write(sequence.json())

    def _onMDAFrame(self, img: np.ndarray, event: MDASequence) -> None:
        index = self.event_to_index(self._axis_order, event)
        name = (
            "_".join(
                [
                    self._axis_order[i] + f"{index[i]}".zfill(3)
                    for i in range(len(index))
                ]
            )
            + ".tiff"
        )
        tifffile.imwrite(self._path / name, img)


class ZarrWriter(BaseWriter):
    def __init__(
        self,
        store_name: Union[str, Path],
        img_shape: Tuple[int, int],
        dtype: npt.DTypeLike,
        core: CMMCorePlus = None,
    ):
        """
        Parameters
        ----------
        store_name : str
            Should accept .format(run=INT)
        img_shape : (int, int)
        dtype : numpy dtype
        core : CMMCorePlus, optional
            If not given the current core instance will be used.
        """
        if zarr is None:
            raise ValueError(
                "This writer requires zarr to be installed. Try: `pip install zarr`"
            )
        super().__init__(core)

        self._store_name = str(store_name)
        self._img_shape = img_shape
        self._dtype = dtype

    def _onMDAStarted(self, sequence: MDASequence):
        self._axis_order = self.sequence_axis_order(sequence)

        name = self.get_unique_folder(self._store_name, suffix=".zarr")
        assert isinstance(name, (Path, str))
        self._z = zarr.open(
            name,
            # self._store_name.format(run=self._run_number),
            mode="w",
            shape=sequence.shape + self._img_shape,
            dtype=self._dtype,
        )
        self._z.attrs["axis_order"] = sequence.axis_order + "yx"
        self._z.attrs["useq-sequence"] = sequence.json()

    def _onMDAFrame(self, img: np.ndarray, event: MDAEvent):
        self._z[self.event_to_index(self._axis_order, event)] = img



from pathlib import Path

from pymmcore_mda_writers import BaseWriter


def test_unique_folder(tmp_path: Path):
    base_folder = tmp_path / "data/run"
    unique = BaseWriter.get_unique_folder
    for i in range(3):
        print("here: ", i)
        print(unique(base_folder, create=True))
    unique(base_folder)

    data_folders = list(map(str, set(tmp_path.glob("data/run_*"))))
    print(data_folders)
    expected = [
        "run_1",
        "run_2",
        "run_3",
    ]
    for e in expected:
        assert str(tmp_path / "data" / e) in data_folders



"""
TiffWriter

This writer will write individual TIFF files
with names of the form t000_p000_c000_z000.tiff
"""

import json

import tifffile
from pymmcore_plus import CMMCorePlus
from useq import MDASequence

from pymmcore_mda_writers import SimpleMultiFileTiffWriter

core = CMMCorePlus.instance()
core.loadSystemConfiguration()

mda = MDASequence(
    time_plan={"interval": 0.1, "loops": 2},
    stage_positions=[(1, 1, 1)],
    z_plan={"range": 3, "step": 1},
    channels=[{"config": "DAPI", "exposure": 1}],
)

# This will automatically get the same core instance we used above
# if you are not using the core singleton make sure to pass core to
# the writer init.
writer = SimpleMultiFileTiffWriter("data/tiff_writer_example/run")


th = core.run_mda(mda)

# wait until the MDA finishes
th.join()


# You can load the saved images using tifffile
print(tifffile.imread("data/tiff_writer_example/run_1/t000_p000_c000_z000.tiff"))

# you can also load the MDASequence from `useq-sequence.json` file
# which is saved a json file in the same directory as the images
with open("data/tiff_writer_example/run_1/useq-sequence.json") as f:
    saved_sequence = json.load(f)

# reconstruct as an MDASequence
saved_sequence_as_useq = MDASequence(**saved_sequence)



import json
from pathlib import Path
from typing import TYPE_CHECKING
from unittest.mock import MagicMock, patch

import numpy as np
import pytest
import zarr
from pymmcore_plus import CMMCorePlus
from pymmcore_plus.mda import MDAEngine
from useq import MDASequence

from pymmcore_mda_writers import BaseWriter, SimpleMultiFileTiffWriter, ZarrWriter

if TYPE_CHECKING:
    from pytestqt.qtbot import QtBot


@pytest.fixture
def core() -> CMMCorePlus:
    mmc = CMMCorePlus.instance()
    if len(mmc.getLoadedDevices()) < 2:
        mmc.loadSystemConfiguration(str(Path(__file__).parent / "test-config.cfg"))
    return mmc


def test_engine_registration(core: CMMCorePlus, tmp_path: Path, qtbot: "QtBot"):
    mda = MDASequence(
        metadata={"blah": "blah blah blah"},
        stage_positions=[(1, 1, 1)],
        z_plan={"range": 3, "step": 1},
        channels=[{"config": "DAPI", "exposure": 1}],
    )

    writer = ZarrWriter(  # noqa
        tmp_path / "zarr_data", (512, 512), dtype=np.uint16, core=core
    )
    new_engine = MDAEngine(core)
    with qtbot.waitSignal(core.events.mdaEngineRegistered):
        core.register_mda_engine(new_engine)
    with qtbot.waitSignal(core.mda.events.sequenceFinished):
        core.run_mda(mda)
    with qtbot.waitSignal(core.mda.events.sequenceFinished):
        core.run_mda(mda)
    run1 = zarr.open(tmp_path / "zarr_data_1.zarr")
    arr1 = np.asarray(run1)
    run2 = zarr.open(tmp_path / "zarr_data_2.zarr")
    arr2 = np.asarray(run2)
    assert arr1.shape == (1, 1, 4, 512, 512)
    assert arr2.shape == (1, 1, 4, 512, 512)
    for i in range(4):
        assert not np.all(arr1[0, 0, i] == 0)
        assert not np.all(arr2[0, 0, i] == 0)
    attrs = run2.attrs.asdict()
    assert "tpczyx" == attrs["axis_order"]
    assert mda == MDASequence(**json.loads(attrs["useq-sequence"]))


def test_tiff_writer(core: CMMCorePlus, tmp_path: Path, qtbot: "QtBot"):
    mda = MDASequence(
        metadata={"blah": "blah blah blah"},
        time_plan={"interval": 0.1, "loops": 2},
        stage_positions=[(1, 1, 1)],
        z_plan={"range": 3, "step": 1},
        channels=[{"config": "DAPI", "exposure": 1}],
    )
    writer = SimpleMultiFileTiffWriter(str(tmp_path / "mda_data"), core=core)  # noqa

    # run twice to check that we aren't overwriting files
    with qtbot.waitSignal(core.mda.events.sequenceFinished):
        core.run_mda(mda)
    with qtbot.waitSignal(core.mda.events.sequenceFinished):
        core.run_mda(mda)

    # check that the correct folders/files were generated
    data_folders = set(tmp_path.glob("mda_data*"))
    assert {tmp_path / "mda_data_1", tmp_path / "mda_data_2"}.issubset(
        set(data_folders)
    )
    expected = [
        Path("t000_p000_c000_z000.tiff"),
        Path("t001_p000_c000_z000.tiff"),
        Path("t001_p000_c000_z002.tiff"),
        Path("t001_p000_c000_z001.tiff"),
        Path("t000_p000_c000_z001.tiff"),
        Path("t001_p000_c000_z003.tiff"),
        Path("t000_p000_c000_z002.tiff"),
        Path("t000_p000_c000_z003.tiff"),
    ]
    actual_1 = list((tmp_path / "mda_data_1").glob("*"))
    actual_2 = list((tmp_path / "mda_data_2").glob("*"))
    for e in expected:
        assert tmp_path / "mda_data_1" / e in actual_1
        assert tmp_path / "mda_data_2" / e in actual_2
    with open(tmp_path / "mda_data_1" / "useq-sequence.json") as f:
        seq = MDASequence(**json.load(f))
    assert seq == mda


def test_missing_deps():
    with patch("pymmcore_mda_writers._writers.tifffile", None):
        with pytest.raises(ValueError) as e:
            SimpleMultiFileTiffWriter("blarg")
        assert "requires tifffile to be installed" in str(e)
    with patch("pymmcore_mda_writers._writers.zarr", None):
        with pytest.raises(ValueError) as e:
            ZarrWriter("blarg", (512, 512), np.uint16)
        assert "requires zarr to be installed" in str(e)


def test_deregistration(core: CMMCorePlus, qtbot: "QtBot"):
    mda = MDASequence(
        stage_positions=[(1, 1, 1)],
        time_plan={"interval": 0.1, "loops": 3},
        channels=[{"config": "DAPI", "exposure": 1}],
    )

    writer = BaseWriter(core)
    writer._disconnect(core.mda)
    writer._onMDAFrame = MagicMock()
    writer._on_mda_engine_registered(core.mda, None)
    new_engine = MDAEngine(core)
    with qtbot.waitSignal(core.events.mdaEngineRegistered):
        core.register_mda_engine(new_engine)
    with qtbot.waitSignal(core.mda.events.sequenceFinished):
        core.run_mda(mda)
    writer.disconnect()
    with qtbot.waitSignal(core.mda.events.sequenceFinished):
        core.run_mda(mda)
    assert writer._onMDAFrame.call_count == 3
